import numpy as np    # arrays and stuff
import pylab as pl    # for plotting
import seaborn as sns # for pretty statistics plotting
from scipy.stats import kstest
from scipy.stats import beta
from scipy.misc import logsumexp # for numerically stable sums of small positive values
import pandas as pd   # for reading csv files in convenient format
from scipy.special import erfinv
from mpmath import mp
mp.dps = 50


dataframe2 = pd.read_csv('./rare_event_is_data.csv')
dataframe = pd.read_csv('./rare_event_data_20171010.csv')

dataframe_naive = pd.read_csv('./paths_criticality_naive.csv')

crits_naive = dataframe_naive['critValue']
porig = dataframe['orig_path_probability']
qprop = dataframe['guide_path_probability']
feval = dataframe['CrashFlag_0.5m']
crits = dataframe['critValue']
weights = porig / qprop
logweights = np.log(porig) - np.log(qprop)
critmask  = feval>0.1
confidence_level = 0.99

sns.set_context('talk')
sns.set_style('ticks')
fig, ax = pl.subplots()
ax.set_yscale('log')
bins = np.linspace(0.,1., 12)

def clopper_pearson_conf(nsucc, nall, alpha=0.05):
    """calculates the confidence interval based on the number of successes
    and overall trials using the clopper pearson exact confidence interval"""
    lo = beta.ppf(alpha/2., nsucc, nall-nsucc+1)
    hi = beta.ppf(1. - alpha/2., nsucc+1, nall-nsucc)
    if nsucc ==0:
        lo = 0.
        hi = 1.- (alpha/2.)**(1./nall)
    return np.array([lo,hi])

cdf_naive = np.array([np.sum(1.*(crits_naive < b))/len(crits_naive) for b in bins])
cdf_naive_conf = np.array([clopper_pearson_conf(int(c), int(len(crits_naive)), 1.-confidence_level)
                  for c in (cdf_naive*len(crits_naive))])
ax.plot(bins, cdf_naive, color='k',
        label=r'empirical cdf - direct sampling $\pm $ %g CI'%(confidence_level))
ax.fill_between(bins, cdf_naive_conf[:,0],
                cdf_naive_conf[:,1], alpha=0.5)
ax.plot(bins, cdf_naive, color='k')
pl.grid()
pl.xlabel('criticality value [m]')
pl.ylabel(r'cumulative probability $P(c \leq x)$')
pl.legend()
sns.despine(offset=10)
pl.savefig('cumulative_critical_distribution_naive_sampling_confidence_%f.pdf'%(confidence_level), bbox_inches='tight')

from tqdm import tqdm
def logsumexpf(lw):
    la = np.max(lw)
    return np.sum(np.array(map(mp.exp,lw-la))) + la

def compute_bootstrap_confidence_interval(logweights, fs , confidence_level=0.9, nbootstrap=1000):
    resampled = np.zeros(nbootstrap)
    for n in tqdm(range(nbootstrap)):
        if nbootstrap==1:
            ind = range(len(logweights))
        else:
            ind = np.random.randint(len(logweights), size=len(logweights))
        #lws = (np.log(porig)/70. - np.log((np.random.rand(len(porig)))*0.01))*70. - np.log(qprop)
        #po = porig**(1./70.)
        #po = np.array(map(lambda x:mp.mpf(x)**(1./70), porig))
        #lws = np.array([mp.log(max(p, 1e-50)) for p in (po + np.array(map(mp.mpf, (np.random.rand(len(porig))-0.5)* 0.01)))])*70. - np.array(map(lambda x: mp.log(mp.mpf(x)), qprop))
        lws = np.array([mp.log(max(mp.mpf(p),1e-80)) for p in porig]) - np.array(map(lambda x: mp.log(mp.mpf(x)), qprop))
        f = fs[ind]
        lws = lws[ind]
        if len(f[f>0.001])>0:
            resampled[n] = mp.exp(logsumexpf(lws[f>0.01]) - logsumexpf(lws))
        r = resampled[n]
        # err = mp.exp(logsumexpf(np.array([2.*l + 2.*mp.log(np.abs(mp.mpf(fe)-r)+1e-40 )
        #                                  for l,fe in zip(lws - logsumexpf(lws), f)])))
        # err = mp.sqrt(err/len(ind))*(mp.sqrt(2.)*erfinv(1.-confidence_level))
        # #print "-: " , r-err , ", +: ", r+err
    #     resampled[n] = np.float(max([r-err,0]))
    #     resampled[n+nbootstrap]= np.float(r+err)
    resampled.sort()
    i1 = int(np.floor((1. - confidence_level) * nbootstrap))
    i1 = max([0, i1])
    i2 = int(np.ceil((confidence_level) * nbootstrap))
    i2 = min([len(resampled) - 1, i2])
    return np.array([resampled[i1],
                     resampled[i2]])

fig, ax = pl.subplots(figsize=(12, 8))
ax.set_yscale('log')



ax.fill_between(bins, cdf_naive_conf[:,0], cdf_naive_conf[:,1] , color='b', alpha=0.5)
#print crits
#print " crits*crits<b: ", [[c if c<=b else 0 for c in crits] for b in bins]
cdf_conf = []

from parallel import mapp
from shmarray import zeros
nbootstrap = 2000
cdf_conf = zeros((len(bins),2))

def f(arg):
    i,b = arg

    l,u = compute_bootstrap_confidence_interval(logweights, np.array([c if c<=b else 0. for c in crits]),
                                                confidence_level=confidence_level, nbootstrap=nbootstrap)
    cdf_conf[i,0] = l
    cdf_conf[i,1] = u

args = []
for i,b in enumerate(bins):
    args.append((i,b))
mapp(f, args)

# for b in bins:
#     cdf_conf.append(compute_bootstrap_confidence_interval(logweights,
#                                                            np.array([c if c<=b else 0. for c in crits]),
#                                                            confidence_level=confidence_level,
#                                                            nbootstrap=nbootstrap))
cdf_conf = np.array(cdf_conf)
#cdf_conf = np.array([compute_bootstrap_confidence_interval(logweights,
#                                                           np.array([c if c<=b else 0. for c in crits]),
#                                                           confidence_level=confidence_level,
#                                                           nbootstrap=500)  for b in bins])

cdf = cdf_conf.mean(axis=1)
mean_zsampling = np.array([compute_bootstrap_confidence_interval(logweights, np.array([c if c<=b else 0. for c in crits]), nbootstrap=1)[0] for b in bins])


#cdf_err = np.array([np.exp(logsumexp(np.array([2.*lw + 2.*np.log(np.abs(f-c)+1e-5)
#                                               for lw,f,c in zip(logweights,crits*(crits<b),cdf)])))
#                    for b in bins])
#ax.fill_between(bins,
#                cdf - np.sqrt(cdf_err/len(cdf_err))*(np.sqrt(2)*erfinv(1.-confidence_level)),
#                cdf +np.sqrt(cdf_err/len(cdf_err))*(np.sqrt(2)*erfinv(1.-confidence_level)),
#               color='g', alpha=0.5)
ax.fill_between(bins,
                cdf_conf[:,0],
                cdf_conf[:,1],
               color='g', alpha=0.5)

ax.plot(bins, cdf_naive, color='b', label=r'emp. cdf - direct sampling $\pm $ %g CI'%(confidence_level))
#ax.plot(bins, cdf, color='y', label=r'mean confidence intervall - guided sampling $\pm $ %g CI'%(confidence_level))
ax.plot(bins, mean_zsampling, color='g', label=r'emp. cdf - guided sampling $\pm $ %g CI'%(confidence_level))
ax.set_yticks(np.logspace(-15, -1, 15))
pl.grid()
pl.xlabel('critical value [m]')
pl.ylabel('cumulative probability $P(c \leq x)$')
pl.ylim(1e-15, 1e-1)
pl.xlim(0, 1)
pl.legend(frameon=False)
sns.despine(offset=10)

pl.savefig('cumulative_critical_distribution_confidence_%f_zsampling_bootstrap_samples_%d.pdf'%(confidence_level, nbootstrap),bbox_inches='tight')

#if __name__ == "__main__":
